/*
 * Copyright 2014-2024  the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package dev.jeka.core.api.system;

import dev.jeka.core.api.utils.JkUtilsAssert;
import dev.jeka.core.api.utils.JkUtilsReflect;
import dev.jeka.core.api.utils.JkUtilsSystem;

import java.io.ByteArrayOutputStream;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * Represents a handler for a running process.
 * This class provides methods to interact with the process.
 */
public class JkProcHandler {

    private final Process process;

    private final ByteArrayOutputStream outputStream;

    JkProcHandler(Process process, ByteArrayOutputStream outputStream) {
        this.process = process;
        this.outputStream = outputStream;
    }

    /**
     * Returns the process associated with this handler.
     */
    public Process getProcess() {
        return process;
    }

    /**
     * Returns true if the executed process has collected output.
     * <p>
     * To collect output, process has to be configured explicitly by invoking {@link JkAbstractProcess#setCollectStdout(boolean)}.
     */
    public boolean hasCollectedOutput() {
        return outputStream != null;
    }

    /**
     * Returns the output generated by the process.
     *
     * @throws IllegalStateException if the process has not collected output
     */
    public String getOutput() {
        JkUtilsAssert.state(outputStream != null, JkProcResult.NO_COLLECT_MSG);
        return outputStream.toString();
    }

    /**
     * Returns the output generated by the executed process as a list of strings.
     */
    public List<String> getOutputMultiline() {
        return  Arrays.asList(getOutput().split("\\r?\n"));
    }

    public int waitFor() {
        try {
            return process.waitFor();
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    public boolean waitFor(long timeout, TimeUnit timeUnit) {
        try {
            return process.waitFor(timeout, timeUnit);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }



}
